import errno
import json
import os
from locale import getpreferredencoding
from pathlib import Path
from typing import Dict, Optional

from filelock import SoftFileLock
from pydantic import BaseSettings


class Settings(BaseSettings):
    API_V1_STR: str = "/api/v1"
    LOCAL_STACK: bool = False
    E2E: bool = False
    CI_RUNNER_NAME: Optional[str] = None

    VERBOSE: bool = False
    DEBUG: bool = False
    WEB_URL: str = "https://app.taktile.com"
    TAKTILE_API_URL: str = "https://taktile-api.taktile.com"
    DEPLOYMENT_API_URL: str = "https://deployment-api.taktile.com"

    TAKTILE_CONFIG_FILE: Path = (
        Path("~/.config/tktl/config.json").expanduser().absolute()
    )
    HELP_HEADERS_COLOR: str = "yellow"
    HELP_OPTIONS_COLOR: str = "green"
    USE_CONSOLE_COLORS: bool = True
    HELP_COLORS_DICT: Dict = {
        "help_headers_color": "yellow",
        "help_options_color": "green",
    }

    LOCAL_ARROW_ENDPOINT: str = "grpc://127.0.0.1:5005"
    LOCAL_REST_ENDPOINT: str = "http://127.0.0.1:8080"
    ARROW_BATCH_MB: int = 50
    PARQUET_BATCH_DEFAULT_ROWS_READ: int = int(1e6)

    FUTURE_ENDPOINTS: bool = True
    TAKTILE_PROFILING_VERSION: Optional[str] = os.environ.get(
        "TAKTILE_PROFILING_VERSION"
    )

    class Config:
        case_sensitive = True


settings = Settings()


def set_api_key(api_key: Optional[str]):
    """set_api_key
    Stores a specified API key in the config file at `TAKTILE_CONFIG_FILE`

    Parameters
    ----------
    api_key : str
        API key to set in the config file
    """

    encoding = getpreferredencoding(False)
    config_path = settings.TAKTILE_CONFIG_FILE
    try:
        os.makedirs(config_path.parent)
    except OSError as exc:
        if exc.errno != errno.EEXIST:
            raise

    with SoftFileLock(f"{config_path}.lock", timeout=1):
        if config_path.exists():
            with config_path.open("r", encoding=encoding) as config_file:
                config = json.load(config_file)
        else:
            config = {}

        if api_key is None:
            if "api-key" in config:
                del config["api-key"]
        else:
            config["api-key"] = api_key

        with config_path.open("w", encoding=encoding) as config_file:
            json.dump(config, config_file, indent=2, sort_keys=True)
            config_file.write("\n")



import abc
import pathlib
import typing as t

import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from fastapi import Response
from pydantic.main import ModelMetaclass
from taktile_types.enums.endpoint import EndpointKinds, ProfileKinds

XType = t.Union[
    pd.Series, pd.DataFrame, np.ndarray, ModelMetaclass, t.Any, pathlib.Path, None
]
YType = XType


class Endpoint(abc.ABC):
    kind: EndpointKinds

    def __init__(
        self,
        name: str,
        position: int,
        func: t.Callable[..., t.Coroutine[t.Any, t.Any, Response]],
        X: XType = None,
        y: YType = None,
        profile_columns: t.Optional[t.List[str]] = None,
        profile: t.Optional[ProfileKinds] = None,
        track_inputs: t.Optional[t.List[str]] = None,
        track_outputs: t.Optional[t.List[str]] = None,
        **kwargs,
    ):
        self._kwargs = kwargs
        self._name = name
        self._position = position
        self._func = func
        self._X = X
        self._y = y
        self._profile_columns = profile_columns
        self._profile_kind = profile
        self._tags = kwargs.get("tags", [])
        self._track_inputs = track_inputs
        self._track_outputs = track_outputs

    @property
    def kwargs(self):
        return self._kwargs

    @property
    def name(self):
        return self._name

    @property
    def position(self):
        return self._position

    @property
    def func(self):
        return self._func

    @property
    def X(self):
        return self._X

    @property
    def y(self):
        return self._y

    @property
    def profile_kind(self):
        return self._profile_kind

    @property
    def profile_columns(self):
        return self._profile_columns

    @property
    def input_names(self):
        return []

    @property
    def output_names(self):
        return ""

    @property
    def tags(self) -> t.List[str]:
        return self._tags

    @property
    def track_inputs(self) -> t.Optional[t.List[str]]:
        return self._track_inputs

    @property
    def track_outputs(self) -> t.Optional[t.List[str]]:
        return self._track_outputs

    @staticmethod
    @abc.abstractmethod
    def supported(
        *,
        X: XType = None,
        y: YType = None,
        profile: t.Optional[str] = None,
    ) -> bool:
        """supported.
        Given the parameters, is this type of endpoint supported?

        Parameters
        ----------
        X : XType
        y : yType
        profile : t.Optional[str]
            the paramters

        Returns
        -------
        bool - true if supported, false else

        """


TaskType = t.Union[t.Tuple, t.Callable]


class EndpointResponse(t.NamedTuple):
    return_value: YType
    tasks: t.Sequence[TaskType] = []



import logging
import time
import typing as t
from functools import singledispatch, wraps
from operator import attrgetter

import pandas as pd  # type: ignore
import pydantic
import requests
from fastapi import Request
from taktile_types.enums.monitor import MonitorSourceType, MonitorType
from taktile_types.schemas.monitor import MonitorDataV2, MonitoringPayload

from tktl.core.exceptions import EndpointInstantiationError
from tktl.future.registration.endpoints.abc import EndpointResponse, XType, YType

from .settings import settings

logger = logging.getLogger(__name__)


class Tracker:
    """Tracker."""

    def __init__(self, request: Request, endpoint_name: str):
        """__init__.

        Parameters
        ----------
        request : Request
            request object of the current request life cycle
        endpoint_name : str
            endpoint's name
        """
        self._values: t.List[MonitorDataV2] = []
        self._request: Request = request
        self._endpoint_name = endpoint_name

    def log_categorical(
        self,
        key: str,
        value: t.Union[int, float, None, t.Sequence[t.Union[int, float, None]]],
        source_type: MonitorSourceType = MonitorSourceType.CUSTOM,
    ):
        """log_categorical logs a categorical value. The differentiation
        between numerical and categorical is important when binning the data.

        Parameters
        ----------
        name : str
            key of the datim
        value : t.Union[int, float, None, t.Sequence[t.Union[int, float, None]]]
            value(s) of the datum
        """
        if not isinstance(value, list):
            value = [value]  # type: ignore

        value = [v if not pd.isnull(v) else None for v in value]

        self._values.append(
            MonitorDataV2(
                value=value,
                type=MonitorType.CATEGORY,
                name=key,
                source_type=source_type,
            )
        )

    def log_numerical(
        self,
        key: str,
        value: t.Union[int, float, None, t.Sequence[t.Union[int, float, None]]],
        source_type: MonitorSourceType = MonitorSourceType.CUSTOM,
    ):
        """log_numerical logs a numerical value. The differentiation between
        numerical and categorical is important when binning the data.

        Parameters
        ----------
        key : str
            key of the datum
        value : t.Union[int, float, None, t.Sequence[t.Union[int, float, None]]]
            value(s) of the datum
        """
        if not isinstance(value, list):
            value = [value]  # type: ignore

        value = [v if not pd.isnull(v) else None for v in value]
        self._values.append(
            MonitorDataV2(
                value=value, type=MonitorType.NUMERIC, name=key, source_type=source_type
            )
        )

    def finalize_payload(self) -> t.Optional[MonitoringPayload]:
        """finalize_payload finalizes the payload to be sent across the wire.

        Returns
        -------
        t.Optional[MonitoringPayload]

        The return is None if no data has been added during this request life
        cycle.
        """
        if not self._values:
            return None

        return MonitoringPayload(
            version=2,
            git_sha=settings.TAKTILE_GIT_SHA,
            git_ref=settings.TAKTILE_GIT_REF,
            repository_id=settings.REPOSITORY_ID,
            endpoint=self._endpoint_name,
            user_agent=self._request.headers.get("user-agent"),
            timestamp=int(time.time() * 1000),
            data=self._values,
        )


def tracker_inject(
    request: Request,
    endpoint_name: str,
    track_inputs: t.Optional[t.List[str]] = None,
    track_outputs: t.Optional[t.List[str]] = None,
):
    """Use tracker_inject to inject a Tracker into a function and
    handle data tracking of inputs and outputs.

    Example
    -------

    @create_tracker_inject_wrapper(request)
    def f(x: int, t: Tracker) -> float:
        ...

    is a function f(x: int) -> float


    Parameters
    ----------
    request : Request
        request of the current request life cycle
    endpoint_name : str
        endpoint's name
    track_inputs : t.Optional[t.List[str]]
        input keys to be tracked
    track_outputs : t.Optional[t.List[str]]
        output keys to be tracked
    """

    def inject_tracker_wrapper(func):

        type_hints = t.get_type_hints(func)
        tracker_kwarg = [
            a for (a, b) in type_hints.items() if a != "return" and b == Tracker
        ]

        @wraps(func)
        def wrapper(*args, **kwargs):
            if not tracker_kwarg and track_inputs is None and track_outputs is None:
                # Do this in the beginning to reduce impact on untracked functions
                return func(*args, **kwargs)

            tracker = Tracker(request, endpoint_name)

            if tracker_kwarg:
                kwargs[tracker_kwarg[0]] = tracker
            _track_values(args[0], track_inputs, tracker, MonitorSourceType.INPUT)
            result = func(*args, **kwargs)

            if isinstance(result, EndpointResponse):
                _track_values(
                    result.return_value,
                    track_outputs,
                    tracker,
                    MonitorSourceType.OUTPUT,
                )

            else:
                _track_values(result, track_outputs, tracker, MonitorSourceType.OUTPUT)

            payload = tracker.finalize_payload()

            if payload is None:
                return result

            if isinstance(result, EndpointResponse):
                result.tasks.append(lambda: _send_payload(payload))
                return result

            else:
                return EndpointResponse(
                    return_value=result, tasks=[lambda: _send_payload(payload)]
                )

        return wrapper

    return inject_tracker_wrapper


def _send_payload(payload: MonitoringPayload):
    if settings.TKTL_MONITORING_ENDPOINT is not None:
        requests.post(
            settings.TKTL_MONITORING_ENDPOINT,
            data=payload.json(),
            headers={"Content-Type": "application/json"},
            timeout=3,
        )
    else:
        print(payload.json())  # noqa


def _pandas_dtype_to_monitor_type(t: pd.Series.dtype) -> MonitorType:
    if t.dtype.kind in ["i", "f"]:
        return MonitorType.NUMERIC
    return MonitorType.CATEGORY


def _track_values(
    value: t.Union[XType, YType],
    keys: t.Optional[t.List[str]],
    tracker: Tracker,
    source_type: MonitorSourceType,
) -> None:
    if keys is None:
        return

    for key in keys:
        tracking_data = extract_monitor_element(value, key=key)
        if tracking_data is not None:

            if tracking_data[1] == MonitorType.CATEGORY:
                tracker.log_categorical(key, tracking_data[0], source_type)  # type: ignore
            elif tracking_data[1] == MonitorType.NUMERIC:
                tracker.log_numerical(key, tracking_data[0], source_type)  # type: ignore


@singledispatch
def extract_monitor_element(
    value: t.Union[XType, YType], *, key: str
) -> t.Optional[t.Tuple[t.Sequence[t.Union[str, float, int]], MonitorType]]:
    return None


@extract_monitor_element.register
def _extract_df(
    value: pd.DataFrame, key: str
) -> t.Optional[t.Tuple[t.Sequence[t.Union[str, float, int]], MonitorType]]:
    column = value[key]
    return extract_monitor_element(column, key)


@extract_monitor_element.register
def _extract_series(
    value: pd.Series, key: str
) -> t.Optional[t.Tuple[t.Sequence[t.Union[str, float, int]], MonitorType]]:
    try:
        type_ = _pandas_dtype_to_monitor_type(value)
        python_list = value.tolist()

        if python_list:
            if value.dtype.kind == "M":
                python_list = [
                    int(x.timestamp()) if not pd.isnull(x) else None
                    for x in python_list
                ]
            return python_list, type_

        return None
    except KeyError as exc:
        logger.warning("KeyError while extracting value from series: %s", str(exc))
        return None


PYDANTIC_TYPE_TRANSLATION = {
    "integer": MonitorType.NUMERIC,
    "boolean": MonitorType.CATEGORY,
    "number": MonitorType.NUMERIC,
    "string": MonitorType.CATEGORY,
}


def _fetch_reference(schema: t.Dict[str, t.Any], reference: str):
    """
    Follow json references. Pydantic creates these in the form
    `{'$ref': '#/definitions/ReferencedModel'}` and adds all
    referenced definitions at the schema's top level, e.g.
    {
      'definitions': {
        'ReferencedModel':{
          'description': 'An enumeration.',
          'enum': ['a', 'b', 'c'],
          'title': 'ReferencedModel',
          'type': 'string'
        }
      }
    }
    """
    key, _, rest = reference.partition("/")
    if key == "#":  # special top-level key - ignore
        return _fetch_reference(schema, rest)
    if rest:
        return _fetch_reference(schema[key], rest)
    return schema[key]


@extract_monitor_element.register
def _extract_pydantic(
    value: pydantic.BaseModel, key: str
) -> t.Optional[t.Tuple[t.Sequence[t.Union[str, float, int]], MonitorType]]:

    parent_key, _, child_key = key.rpartition(".")

    try:
        parent = value if not parent_key else attrgetter(parent_key)(value)
        item = getattr(parent, child_key)

        child_schema = parent.schema()["properties"][child_key]
        # resolve reference types
        # these are saved under 'definitions' in the top level schema
        # See https://pydantic-docs.helpmanual.io/usage/schema/ for details
        if "$ref" in child_schema:
            child_schema = _fetch_reference(value.schema(), child_schema["$ref"])

        if "enum" in child_schema:
            # depending on whether the pydantic attribute `use_enum_values`
            # is set, we'll get either an enum or its value here
            return ([getattr(item, "value", item)], MonitorType.CATEGORY)

        type_ = PYDANTIC_TYPE_TRANSLATION[child_schema["type"]]
        return ([item], type_)

    except (KeyError, AttributeError) as exc:
        logger.warning("KeyError while extracting pydantic value: %s", str(exc))
        return None


def verify_tracking_config(value: t.Union[XType, YType], keys: t.List[str]):
    if isinstance(value, pd.Series):
        if len(keys) >= 2:
            raise EndpointInstantiationError(
                "Trying to track multiple values of a series"
            )
    if isinstance(value, pd.DataFrame):
        tracked = set(keys)
        given = set(value.columns.tolist())
        if not tracked.issubset(given):
            raise EndpointInstantiationError(
                f"Couldn't find the following tracking keys: {tracked.difference(given)}"
            )



from functools import wraps

import requests
from pydantic.types import UUID4
from taktile_client.config import collect_api_key

from tktl.core.config import set_api_key, settings
from tktl.core.loggers import LOG
from tktl.core.strings import CLIStrings, HeaderStrings


def login(api_key=None):
    """

    Parameters
    ----------
    api_key

    Returns
    -------

    """
    if api_key is None:
        LOG.error("Authentication failed: no key provided")
        return False
    try:
        UUID4(api_key)
    except ValueError:
        LOG.error("Authentication failed: Key format is invalid")
        return False
    set_api_key(api_key)
    return True


def logout():
    set_api_key(None)
    return True


def validate_key(suppress=False, api_key=None):
    if not api_key:
        api_key = collect_api_key()
    if not api_key:
        LOG.error(CLIStrings.AUTH_ERROR_MISSING_KEY)
        return False

    r = requests.get(
        f"{settings.TAKTILE_API_URL}{settings.API_V1_STR}/users/me",
        headers={"Accept": HeaderStrings.APPLICATION_JSON, "X-Api-Key": api_key},
    )
    try:
        r.raise_for_status()
    except requests.exceptions.RequestException:
        LOG.error(f'Request failed: {r.json()["detail"]}')
        return False

    if not suppress:
        LOG.log("Login successful!", color="green")
    return True


def validate_decorator(func):
    @wraps(func)
    def wrapped_validation(*args, **kwargs):
        if not validate_key(suppress=True):
            return
        return func(*args, **kwargs)

    return wrapped_validation



from pydantic import BaseModel


class Model(BaseModel):
    """
    Model class

    :param str id:
    :param str name:
    :param str project_id:
    :param str experiment_id:
    :param list tags:
    :param str model_type:
    :param str url:
    :param str model_path:
    :param str deployment_state:
    :param str summary:
    :param str detail:
    """

    id: str
    name: str
    project_id: str
    experiment_id: str
    tags: list
    model_type: str
    url: str
    model_path: str
    deployment_state: str
    summary: dict
    detail: dict
    notes: str


class ModelFile(BaseModel):
    """
    Model file

    :param str file: file path
    :param str url: Url with AWS key
    :param int size: File size in bytes
    """

    file: str
    url: str
    size: int



from typing import List, Union

from pydantic import BaseModel

from tktl.core.t import ProjectAssetSourceT, ProjectAssetT


class ProjectContentsBase(BaseModel):
    type: str
    name: str
    path: str


class ProjectFile(ProjectContentsBase):
    type: str = "file"


class ProjectFileWithContent(ProjectContentsBase):
    type: str = "file"
    content: str


class ProjectDirectory(ProjectContentsBase):
    type: str = "dir"


class ProjectAsset(ProjectContentsBase):
    calculated_sha: str
    kind: ProjectAssetT
    source: ProjectAssetSourceT
    requires_download: bool


class TktlYamlConfigValidationError(BaseModel):
    loc: str
    msg: str


class ProjectValidationOutput(BaseModel):
    title: str
    summary: str
    text: str


ProjectContentSingleItemT = Union[ProjectFile, ProjectDirectory]
ProjectContentMultiItemT = List[ProjectContentSingleItemT]
ProjectContentT = Union[
    ProjectContentSingleItemT, ProjectContentMultiItemT, ProjectFileWithContent
]



import typing as t

from taktile_types.enums.endpoint import EndpointKinds

from tktl.future.monitor import verify_tracking_config

from .abc import Endpoint, XType, YType


class GenericEndpoint(Endpoint):
    kind: EndpointKinds = EndpointKinds.GENERIC

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.track_inputs:
            verify_tracking_config(self.X, self.track_inputs)
        if self.track_outputs:
            verify_tracking_config(self.y, self.track_outputs)

    @staticmethod
    def supported(
        *,
        X: XType = None,
        y: YType = None,
        profile: t.Optional[str] = None,
    ) -> bool:
        return profile is None and X is None and y is None



import pathlib
import re
from typing import Any, Callable, Dict, List, Optional, Sequence, Union, cast

from tktl.core.exceptions import TaktileSdkError
from tktl.core.loggers import LOG
from tktl.core.schemas.repository import (
    Endpoint,
    Repository,
    RepositoryDeployment,
    Resources,
)


def default_comp(x: Any, y: Any):
    return x == y


def branch_comp(x: str, y: str):
    return x == y or y == f"refs/heads/{x}"


def is_in_sequence_comp(x: Sequence, y: str):
    return y in set(x)


def filter_prop(
    resources: Sequence[Resources],
    prop_name: str,
    value: Any,
    key: Callable = default_comp,
):
    if value is not None:
        return [d for d in resources if key(value, getattr(d, prop_name))]
    return resources


def filter_deployments(
    deployments: List[RepositoryDeployment],
    git_hash: Optional[str] = None,
    branch_name: Optional[str] = None,
    status_name: Union[str, List[str], None] = None,
) -> List[RepositoryDeployment]:
    for n, v in zip(
        ["commit_hash", "git_ref", "status"], [git_hash, branch_name, status_name]
    ):
        if n == "git_ref":
            comp = branch_comp
        elif n == "commit_hash" and git_hash:
            if len(git_hash) < 40:

                def comp(x: str, y: str):
                    return y.startswith(x)

            else:
                comp = default_comp
        elif n == "status":
            if isinstance(status_name, list):
                comp = is_in_sequence_comp
            else:
                comp = default_comp
        else:
            comp = default_comp
        deployments = filter_prop(deployments, prop_name=n, value=v, key=comp)

    return deployments


def filter_repositories(
    repositories: List[Repository],
    repository_name: Optional[str] = None,
    repository_owner: Optional[str] = None,
) -> List[Repository]:
    for n, v in zip(
        ["repository_name", "repository_owner"], [repository_name, repository_owner]
    ):
        repositories = filter_prop(repositories, prop_name=n, value=v)
    return repositories


def filter_endpoints(
    endpoints: List[Endpoint],
    endpoint_name: Optional[str] = None,
    endpoint_kind: Optional[str] = None,
) -> List[Endpoint]:

    for n, v in zip(["name", "kind"], [endpoint_name, endpoint_kind]):
        endpoints = filter_prop(endpoints, prop_name=n, value=v)
    return endpoints


def get_deployment_from_endpoint_and_branch(
    endpoint_mapping: List[Dict[str, Union[Endpoint, RepositoryDeployment]]],
    endpoint_name: str,
    branch_name: str = None,
) -> Optional[RepositoryDeployment]:
    filtered = [
        e
        for e in endpoint_mapping
        if e and cast(Endpoint, e["endpoint"]).name == endpoint_name
    ]
    if len(filtered) == 0:
        LOG.warning(
            f"No endpoints with name: {endpoint_name} found across all deployed branches"
        )
        return None
    if len(filtered) == 1:
        deployment = cast(RepositoryDeployment, filtered[0]["deployment"])
        if branch_name is not None and not branch_comp(deployment.git_ref, branch_name):
            base_name = pathlib.Path(deployment.git_ref).name
            LOG.warning(
                f"No endpoint {endpoint_name} for branch {branch_name}, but found one for branch: "
                f"{base_name}. Returning that instead"
            )
        return deployment
    else:
        available_branches = set(
            [
                pathlib.Path(cast(RepositoryDeployment, e["deployment"]).git_ref).name
                for e in endpoint_mapping
            ]
        )
        if branch_name is None:
            LOG.warning(
                f"Ambiguous: more than one branch for repository with endpoint: {endpoint_name}. Please "
                f"specify a branch. Available branches: with such endpoint: {', '.join(available_branches)}"
            )
            return None
        branch_filtered = [
            e
            for e in filtered
            if branch_comp(
                cast(RepositoryDeployment, e["deployment"]).git_ref, branch_name
            )
        ]
        if len(branch_filtered) == 0:
            LOG.warning(
                f"No endpoints with name: {endpoint_name} and branch {branch_name} found. Available "
                f"branches with such endpoint: {', '.join(available_branches)}"
            )
            return None
        elif len(branch_filtered) == 1:
            return cast(RepositoryDeployment, branch_filtered[0]["deployment"])
        else:
            # Impossible to have more than one endpoint in the same branch for the same repo with the same name
            raise TaktileSdkError("This should never, ever happen")


def parse_remote_url(remote_url: str):
    match = re.findall(r"(?<=github.com).*", remote_url)
    return pathlib.Path(match[0].strip("/").strip(":")).stem if match else ""



import pathlib
from typing import List, Optional, Tuple, Union

from pydantic import ValidationError
from taktile_types.schemas.repository.config_yaml import TaktileYAMLConfig
from yaml import safe_load

from tktl.core.exceptions import NoContentsFoundException, UserRepoValidationException
from tktl.core.schemas.project import (
    ProjectContentMultiItemT,
    ProjectContentT,
    ProjectDirectory,
    ProjectFile,
    ProjectFileWithContent,
)
from tktl.core.t import (
    RequiredUserProjectPathsT,
    UserProjectFileT,
    UserRepoConfigFileNameT,
)


def get_user_repo_contents(path: str) -> Optional[ProjectContentT]:
    contents: List[Union[ProjectFile, ProjectDirectory]] = []
    for item in pathlib.Path(path).iterdir():
        if item.is_file():
            contents.append(ProjectFile(name=item.name, path=str(item)))
        elif item.is_dir():
            contents.append(ProjectDirectory(name=item.name, path=str(item)))
    return contents


def validate_project_contents(
    path: str,
) -> Tuple[TaktileYAMLConfig, ProjectContentMultiItemT]:
    contents = get_user_repo_contents(path=path)
    config_found = False
    config_file = None
    required_files = RequiredUserProjectPathsT.strictly_required_files()
    required_dirs = RequiredUserProjectPathsT.strictly_required_dirs()

    if not contents:
        raise NoContentsFoundException(
            missing_config=True,
            missing_directories=list(required_dirs),
            missing_files=list(required_files),
        )

    for file in contents:
        if file.type == UserProjectFileT.FILE.value:  # type: ignore
            if file.name in UserRepoConfigFileNameT.set():  # type: ignore
                config_found = True
                config_file = file
            elif file.name in required_files:  # type: ignore
                required_files.remove(file.name)  # type: ignore
        elif file.type == UserProjectFileT.DIRECTORY.value:  # type: ignore
            if file.name in required_dirs:  # type: ignore # type: ignore
                required_dirs.remove(file.name)  # type: ignore

    if not config_found or required_dirs or required_files:
        raise UserRepoValidationException(
            missing_files=list(required_files),
            missing_directories=list(required_dirs),
            missing_config=not config_found,
        )
    config_file_with_content = decode_file_content(config_file)  # type: ignore
    return config_file_with_content, contents  # type: ignore


def decode_file_content(
    file_with_content: ProjectFileWithContent,
) -> Union[TaktileYAMLConfig, ValidationError]:
    with open(file_with_content.path, "r") as yml:
        as_dict = safe_load(yml)
    return TaktileYAMLConfig(**as_dict)



import pathlib
import tempfile

import pandas as pd  # type: ignore
import pytest

from tktl.commands.lazify import lazify_file
from tktl.future.lazy_loading import load_background_data_from_lazified_pq


@pytest.mark.parametrize(
    "df, background_data",
    [
        (
            pd.DataFrame.from_dict(
                {
                    "Integer": pd.Series([1], dtype="int64"),
                    "Float": pd.Series([3.0], dtype="float64"),
                    "Object": pd.Series(["c"], dtype="object"),
                    "Bool": pd.Series([True], dtype="bool"),
                    "Categorical": pd.Series(["a"], dtype="category"),
                    "Datetime": pd.Series(["2020-01-01"], dtype="datetime64[ns]"),
                }
            ),
            pd.DataFrame.from_dict(
                {
                    "Integer": pd.Series([1], dtype="int64"),
                    "Float": pd.Series([3.0], dtype="float64"),
                    "Object": pd.Series(["c"], dtype="object"),
                    "Bool": pd.Series([True], dtype="bool"),
                    "Categorical": pd.Series(["a"], dtype="category"),
                    "Datetime": pd.Series(["2020-01-01"], dtype="datetime64[ns]"),
                }
            ),
        ),
        (
            pd.DataFrame.from_dict(
                {
                    "Integer": pd.Series([1, 3, 5], dtype="int64"),
                    "Float": pd.Series([3.0, 3.5, 4.0], dtype="float64"),
                    "Object": pd.Series(["c", "d", "c"], dtype="object"),
                    "Bool": pd.Series([True, False, False], dtype="bool"),
                    "Categorical": pd.Series(["a", "a", "a"], dtype="category"),
                    "Datetime": pd.Series(
                        ["2020-01-01", "2020-01-01", "2020-01-01"],
                        dtype="datetime64[ns]",
                    ),
                }
            ),
            pd.DataFrame.from_dict(
                {
                    "Integer": pd.Series([3], dtype="int64"),
                    "Float": pd.Series([3.5], dtype="float64"),
                    "Object": pd.Series(["c"], dtype="object"),
                    "Bool": pd.Series([False], dtype="bool"),
                    "Categorical": pd.Series(["a"], dtype="category"),
                    "Datetime": pd.Series(["2020-01-01"], dtype="datetime64[ns]"),
                }
            ),
        ),
    ],
)
def test_file_lazification(df, background_data):

    with tempfile.TemporaryDirectory() as tmp_dir:
        path = pathlib.Path(tmp_dir) / "test.pqt"

        df.to_parquet(path)
        lazify_file(source_path=path, target_path=path, data=None)

        df_loaded = load_background_data_from_lazified_pq(path)

    pd.testing.assert_frame_equal(df_loaded, background_data)



import pathlib

import joblib
import pandas as pd

from tktl.future import Tktl

# instantiate client
client = Tktl()

# load model
model = joblib.load("assets/model.joblib")

# load reference data path
X = pathlib.Path("assets", "loans_test_X.pqt")
y = pathlib.Path("assets", "loans_test_y.pqt")


# specify transformation
@client.endpoint(profile="binary", X=X, y=y)
def repayment(df):
    pred = model.predict_proba(df)[:, 1]
    return pd.Series(pred)



import typing as t
from json import JSONDecodeError

from pydantic import UUID4
from taktile_client.config import collect_api_key
from taktile_client.http_client import API

from tktl.constants import URL_UNAVAILABLE
from tktl.core.clients.utils import (
    filter_deployments,
    filter_endpoints,
    filter_repositories,
)
from tktl.core.config import settings
from tktl.core.exceptions import APIClientException, TaktileSdkError
from tktl.core.loggers import LOG
from tktl.core.schemas.deployment import DeploymentBase
from tktl.core.schemas.repository import (
    Endpoint,
    Repository,
    RepositoryDeployment,
    _format_http_url,
)
from tktl.core.t import DeploymentStatesT
from tktl.core.utils import flatten, lru_cache


class DeploymentApiClient(API):
    SCHEME: str

    def __init__(self, api_key: str):
        super().__init__(api_base=settings.DEPLOYMENT_API_URL, api_key=api_key)
        self._api_key = api_key

    @lru_cache(timeout=50, typed=False)
    def __get_repositories(self) -> t.List[Repository]:
        return self.call(
            verb="get", path=f"{settings.API_V1_STR}/models", model=Repository
        )

    def _get_repositories(self) -> t.List[Repository]:
        repositories = self.__get_repositories()
        if not repositories:
            raise TaktileSdkError(
                "No repositories found on your account, so no resources can be fetched"
            )
        return repositories

    def get_deployments(
        self,
        repository_id: t.Optional[UUID4],
        git_hash: t.Optional[str],
        branch_name: t.Optional[str],
        status_name: t.Optional[str],
        repository_name: t.Optional[str],
        repository_owner: t.Optional[str],
        return_all: bool,
    ) -> t.Optional[t.List[RepositoryDeployment]]:

        repositories = filter_repositories(
            self._get_repositories(),
            repository_name=repository_name,
            repository_owner=repository_owner,
        )

        if repository_id:
            repo = [r for r in repositories if r.id == repository_id]
            if not repo:
                LOG.warning("No repositories with matching id found")
                return None
            else:
                deployments = repo[0].deployments
        else:
            deployments = flatten([r.deployments for r in repositories])

        if return_all:
            filtered_deployments = deployments
        else:
            filtered_deployments = filter_deployments(
                deployments,
                git_hash=git_hash,
                branch_name=branch_name,
                status_name=status_name,
            )

        def get_number_of_endpoints(deployment: RepositoryDeployment):
            try:
                return len(get_endpoints_for_deployment(deployment))
            except (JSONDecodeError, APIClientException):
                return 0

        endpoints_for_deployment = [
            get_number_of_endpoints(deployment) for deployment in filtered_deployments
        ]
        for e, d in zip(endpoints_for_deployment, filtered_deployments):
            d.n_endpoints = e
        return filtered_deployments

    def get_repositories(
        self,
        repository_name: t.Optional[str] = None,
        repository_owner: t.Optional[str] = None,
        return_all: bool = False,
    ) -> t.List[Repository]:
        repositories = self._get_repositories()
        if return_all:
            return repositories
        return filter_repositories(
            repositories,
            repository_name=repository_name,
            repository_owner=repository_owner,
        )

    def get_endpoints(
        self,
        deployment_id: t.Optional[UUID4] = None,
        endpoint_kind: t.Optional[str] = None,
        endpoint_name: t.Optional[str] = None,
        repository_name: t.Optional[str] = None,
        repository_owner: t.Optional[str] = None,
        git_hash: t.Optional[str] = None,
        branch_name: t.Optional[str] = None,
        status_name: t.Optional[str] = None,
        return_all: bool = False,
    ) -> t.List[Endpoint]:
        repositories = filter_repositories(
            self._get_repositories(),
            repository_name=repository_name,
            repository_owner=repository_owner,
        )
        deployments = filter_deployments(
            deployments=flatten([r.deployments for r in repositories]),
            git_hash=git_hash,
            branch_name=branch_name,
            status_name=status_name,
        )

        if deployment_id:
            deployments = [d for d in deployments if d.id == deployment_id]
            if not deployments:
                LOG.warning(
                    f"No endpoints for deployment with id {deployment_id} found"
                )
        endpoints = flatten(
            [get_endpoints_for_deployment(deployment) for deployment in deployments]
        )
        if return_all:
            return endpoints
        return filter_endpoints(
            endpoints, endpoint_kind=endpoint_kind, endpoint_name=endpoint_name
        )

    def get_deployment_by_branch_name(self, repository_name: str, branch_name: str):
        # owner is guaranteed not to have a slash in it so can split on first occurrence
        owner, name = repository_name.split("/", 1)
        repositories = filter_repositories(
            self._get_repositories(),
            repository_name=name,
            repository_owner=owner,
        )
        if not repositories:
            raise TaktileSdkError(f"No repos named {repository_name} found")

        deployments = filter_deployments(
            deployments=flatten([r.deployments for r in repositories]),
            branch_name=branch_name,
            status_name=DeploymentStatesT.list(),
        )
        if not deployments:
            raise TaktileSdkError("No running deployments found")

        if len(deployments) > 1:
            raise TaktileSdkError(
                "More than one deployment for single branch and repo found. "
                "Exiting, as this should not happen"
            )
        return deployments[0]

    def get_endpoint_by_name(
        self, repository_name: str, branch_name: str, endpoint_name: str
    ) -> t.Tuple[RepositoryDeployment, Endpoint]:
        deployment = self.get_deployment_by_branch_name(
            repository_name=repository_name, branch_name=branch_name
        )
        endpoints = get_endpoints_for_deployment(deployment)
        for endpoint in endpoints:
            if endpoint.name == endpoint_name:
                return deployment, endpoint
        raise TaktileSdkError(f"No endpoint named {endpoint_name} found")

    def delete_deployment(self, deployment_id: UUID4) -> DeploymentBase:
        return self.call(
            verb="delete",
            path=f"{settings.API_V1_STR}/deployments/{deployment_id}",
            model=DeploymentBase,
        )


def get_endpoints_for_deployment(deployment: RepositoryDeployment) -> t.List[Endpoint]:
    url = _format_http_url(deployment.public_docs_url, docs=False)
    if url == URL_UNAVAILABLE:
        return []
    client = API(api_base=url, api_key=collect_api_key())
    endpoint_models = client.call(verb="get", path="info", model=Endpoint)
    for endpoint in endpoint_models:
        endpoint.deployment_id = deployment.id
    return endpoint_models



import math
import typing as t

import pandas as pd  # type: ignore
from pyarrow import Table  # type: ignore
from pyarrow.flight import FlightClient  # type: ignore
from pyarrow.flight import ActionType, FlightDescriptor

from tktl.core.config import settings
from tktl.core.loggers import LOG
from tktl.core.serializers import deserialize_arrow, serialize_arrow


def batch_arrow(
    *, client: FlightClient, action: ActionType, X: t.Any, use_input_index: bool = False
):
    table = serialize_arrow(X)
    batch_size, batch_memory = _get_chunk_size(table)
    if not (batch_size and batch_memory):
        return
    descriptor = client.get_flight_info(FlightDescriptor.for_command(action.type))
    writer, reader = client.do_exchange(descriptor.descriptor)  # type: ignore
    LOG.trace(
        f"Initiating prediction request with batches of {batch_size} records of "
        f"~{batch_memory:.2f} MB/batch"
    )
    batches = table.to_batches(max_chunksize=batch_size)
    chunks = []
    schema = None
    with writer:
        writer.begin(table.schema)
        for i, batch in enumerate(batches):
            LOG.trace(f"Prediction for batch {i + 1}/{len(batches)}")
            chunk = _send_batch(
                writer=writer, batch=batch, reader=reader, batch_number=i + 1
            )
            if not chunk:
                continue
            if not schema and chunk.data.schema is not None:
                schema = chunk.data.schema
            chunks.append(chunk.data)
    deserialized = deserialize_arrow(Table.from_batches(chunks, schema))
    if use_input_index:
        input_has_index = isinstance(X, pd.Series) or isinstance(X, pd.DataFrame)
        output_has_index = isinstance(deserialized, pd.Series) or isinstance(
            deserialized, pd.DataFrame
        )
        if not input_has_index or not output_has_index:
            LOG.warning(
                "Inputs or Outputs are not of type series or dataframe, use_input_index has no effect"
            )
        else:
            try:
                deserialized.index = X.index
            except Exception as e:
                LOG.warning(f"Unable to set indexes of output frame: {repr(e)}")
    return deserialized


def _send_batch(writer, batch, reader, batch_number):
    try:
        writer.write_batch(batch)
        return reader.read_chunk()
    except Exception as e:
        LOG.error(
            f"ERROR: performing prediction for batch {batch_number}: {e} "
            f"The predictions from this batch will be missing from the result"
        )
        return None


def _get_chunk_size(sample_table: Table) -> t.Tuple[t.Optional[int], t.Optional[float]]:
    try:
        mem_per_record = sample_table.nbytes / sample_table.num_rows
    except ZeroDivisionError:
        LOG.error(
            "Empty payload received, which is currently not supported for arrow endpoints"
        )
        return None, None
    batch_size = math.ceil(settings.ARROW_BATCH_MB * 1e6 / mem_per_record)
    batch_memory_mb = (batch_size * mem_per_record) / 1e6
    return batch_size, batch_memory_mb



"""Module for user exposed env settings"""
import typing as t

from pydantic import BaseSettings


class Settings(BaseSettings):
    TAKTILE_GIT_SHA: str = "unknown"
    TAKTILE_GIT_REF: str = "unknown"
    REPOSITORY_ID: str = "unknown"
    TKTL_MONITORING_ENDPOINT: t.Optional[str] = None


settings = Settings()



import pathlib
import time
from typing import Dict, Optional

import docker  # type: ignore
import requests
from requests.exceptions import RequestException

from tktl.commands.health import check_grpc_health, check_rest_health
from tktl.core.config import settings
from tktl.core.exceptions import DockerBuildError, MissingDocker
from tktl.core.loggers import LOG

TESTING_DOCKERFILE = "Dockerfile.taktile-cli-testing"
MULTI_STAGE_BUILD_STEP_NAME = "build_step"


class DockerManager:
    def __init__(self, path):
        try:
            self._client = docker.from_env()
            self._path = pathlib.Path(path)
        except docker.errors.DockerException as err:
            raise MissingDocker from err

    def get_docker_file(self) -> str:
        with open(self._path / ".buildfile") as fp:
            return fp.read()

    def stream_logs(self, container) -> None:
        for line in container.logs(stream=True):
            LOG.log(f"> {line.decode()}".strip())

    def remove_image(self, image):
        self._client.images.remove(image, force=True)

    def build_image(
        self,
        dockerfile: str,
        cache: bool,
        prune: bool,
        buildargs: Optional[Dict] = None,
    ) -> str:

        if prune:
            try:
                image = self._client.images.get("taktile-cli-test:latest")
                self._client.images.remove(image.id)
            except docker.errors.ImageNotFound:
                pass

        if not cache:
            LOG.log("Building without Docker cache. This may take some time...")

        LOG.log(
            f"Multistage dockerfile, building target {MULTI_STAGE_BUILD_STEP_NAME}..."
        )
        try:
            image, stream = self._client.images.build(
                path=str(self._path),
                dockerfile=dockerfile,
                tag="taktile-cli-test",
                nocache=not cache,
                target=MULTI_STAGE_BUILD_STEP_NAME,
                buildargs=buildargs,
                forcerm=True,
            )

            for chunk in stream:
                if "stream" in chunk:
                    for line in chunk["stream"].splitlines():
                        LOG.trace("> " + line)

            return image.id
        except docker.errors.BuildError as e:
            for chunk in e.build_log:
                if "stream" in chunk:
                    for line in chunk["stream"].splitlines():
                        LOG.log("> " + line)
            raise DockerBuildError(e.msg)

    def test_import(self, image_id: str, secrets: Optional[Dict[str, str]] = None):
        if secrets is None:
            secrets = {}
        client_name = "tktl" if not settings.FUTURE_ENDPOINTS else "client"
        command = f"python -c 'from src.endpoints import {client_name}'"
        try:
            container = self._client.containers.run(
                image_id,
                command,
                detach=True,
                environment={
                    "TAKTILE_GIT_SHA": "local-run",
                    "TAKTILE_GIT_REF": "local-run",
                    **secrets,
                },
            )
            self.stream_logs(container)

            status = container.wait()
            return status, container
        finally:
            container.remove()

    def test_endpoint_sample(
        self,
        image_id: str,
        secrets: Optional[Dict[str, str]] = None,
    ):
        if secrets is None:
            secrets = {}
        try:
            command = f"python /app/scripts/endpoint_sample_test{'_future' if settings.FUTURE_ENDPOINTS else ''}.py"
            container = self._client.containers.run(
                image_id,
                command,
                detach=True,
                environment={
                    "TAKTILE_GIT_SHA": "local-run",
                    "TAKTILE_GIT_REF": "local-run",
                    **secrets,
                },
            )
            self.stream_logs(container)

            status = container.wait()
            return status, container
        finally:
            container.remove()

    def test_unittest(self, image_id: str, secrets: Optional[Dict[str, str]] = None):
        if secrets is None:
            secrets = {}
        try:
            container = self._client.containers.run(
                image_id,
                "python -m pytest ./user_tests/",
                detach=True,
                environment={
                    "TAKTILE_GIT_SHA": "local-run",
                    "TAKTILE_GIT_REF": "local-run",
                    **secrets,
                },
            )
            self.stream_logs(container)

            status = container.wait()
            return status, container
        finally:
            container.remove()

    def run_rest_container(
        self,
        image_id: str,
        detach: bool = True,
        auth_enabled: bool = True,
        reload: bool = False,
        secrets: Optional[Dict[str, str]] = None,
    ):
        if secrets is None:
            secrets = {}
        assets_path = (self._path / "assets").resolve()
        src_path = (self._path / "src").resolve()
        return self._client.containers.run(
            image_id,
            entrypoint="/start-reload.sh" if reload else "/start-rest.sh",
            detach=detach,
            environment={
                "AUTH_ENABLED": auth_enabled,
                "TAKTILE_GIT_SHA": "local-run",
                "TAKTILE_GIT_REF": "local-run",
                **secrets,
            },
            ports={"80/tcp": 8080},
            remove=True,
            stderr=True,
            stdout=True,
            volumes=[
                f"{assets_path}:/app/assets",
                f"{src_path}:/app/src",
            ]
            if reload
            else [],
        )

    def run_arrow_container(
        self,
        image_id: str,
        detach: bool = True,
        auth_enabled: bool = True,
        secrets: Optional[Dict[str, str]] = None,
    ):
        if secrets is None:
            secrets = {}
        return self._client.containers.run(
            image_id,
            detach=detach,
            entrypoint="/start-flight.sh",
            environment={
                "AUTH_ENABLED": auth_enabled,
                "TAKTILE_GIT_SHA": "local-run",
                "TAKTILE_GIT_REF": "local-run",
                **secrets,
            },
            ports={"5005/tcp": 5005},
            remove=True,
            stderr=True,
            stdout=True,
        )

    def run_containers(
        self,
        image_id: str,
        detach: bool = True,
        auth_enabled: bool = True,
        reload: bool = False,
        secrets: Optional[Dict[str, str]] = None,
    ):
        arrow_container = self.run_arrow_container(
            image_id=image_id, detach=detach, auth_enabled=auth_enabled, secrets=secrets
        )
        rest_container = self.run_rest_container(
            image_id=image_id,
            detach=detach,
            auth_enabled=auth_enabled,
            reload=reload,
            secrets=secrets,
        )
        return arrow_container, rest_container

    def run_profiling_container(self):
        try:
            image_name = f"taktile/taktile-profiler:{_get_profiling_version()}"
            LOG.debug(f"Using Profiling Image {image_name}")
            container = self._client.containers.run(
                image_name,
                entrypoint="profiler profile -l",
                network_mode="host",
                detach=True,
            )
            self.stream_logs(container)
            status = container.wait()
            return status, container
        finally:
            container.remove()

    def run_and_check_health(
        self,
        image_id: str,
        kill_on_success: bool = False,
        auth_enabled: bool = True,
        timeout: int = 7,
        retries: int = 7,
        secrets: Optional[Dict[str, str]] = None,
    ):
        if secrets is None:
            secrets = {}
        arrow_container, rest_container = self.run_containers(
            image_id=image_id, detach=True, auth_enabled=auth_enabled, secrets=secrets
        )

        try:
            while True:
                try:
                    time.sleep(timeout)
                    check_grpc_health()
                    check_rest_health()
                    return arrow_container, rest_container
                except Exception:
                    if retries > 0:
                        retries -= 1
                    else:
                        raise

        finally:
            if kill_on_success:
                arrow_container.kill()
                rest_container.kill()


def _get_profiling_version() -> str:
    if settings.TAKTILE_PROFILING_VERSION:
        return settings.TAKTILE_PROFILING_VERSION
    else:
        try:
            res = requests.get(f"{settings.LOCAL_REST_ENDPOINT}/info").json()
            if isinstance(res, list):  # Not Future
                profiling_version = "latest"
            else:  # Future
                profiling_version = res.get("profiling", "latest")
        except RequestException as e:
            LOG.debug(f"Failed to fetch profiling version from Info {e}")
            profiling_version = "latest"
        # Using py3.7 as it is currently being used in Operator
        # TODO - Replace after finishing https://github.com/taktile-org/taktile-services/issues/2867
        return f"{profiling_version}-py3.7"



from typing import Dict, Optional

from tktl.commands.validate import build_image, get_docker_manager
from tktl.core.loggers import LOG, stream_blocking_logs


def run_container(
    path: str,
    cache: bool,
    prune: bool,
    background: bool,
    auth_enabled: bool,
    reload: bool,
    color_logs: bool,
    secrets: Optional[Dict[str, str]] = None,
):
    dm = get_docker_manager(path=path)
    image = build_image(dm=dm, path=path, cache=cache, prune=prune, secrets=secrets)

    LOG.log("Waiting for service to start...")
    arrow_container, rest_container = dm.run_containers(
        image, detach=True, auth_enabled=auth_enabled, reload=reload, secrets=secrets
    )

    if background:
        return arrow_container, rest_container
    stream_blocking_logs(arrow_container, rest_container, color_logs=color_logs)
